/************************************************************************/
/*  Routines for polygons, derived from polygon_utils in redilib	*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2000 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char sccsid[] = "@(#) $Id: polygon.c,v 1.1 2003/10/29 21:18:17 lombard Exp $";
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "externals.h"
#include "polygon.h"

#define	NEXT_LINE	0
#define	SKIP_COMMENT	1

#define LINELEN	255

#define N 10

static char *fgetline(char *line, int linelen, FILE *fp, int flag);


/************************************************************************/
/*  read_polygon -							*/
/*	Read a polygon in the file, and return structure array of	*/
/*	vertices.							*/
/*	Return number of points read; exit on error.			*/
/************************************************************************/
int read_polygon (FILE *pfp, VERTEX **p_pts)
{
    char line[LINELEN+1];
    int ptnumbuf = 0;
    int npoints = 0;
    VERTEX *ptbuf = NULL;
    VERTEX *pt;
    
    while(fgetline(line,LINELEN,pfp,SKIP_COMMENT) != NULL) {
	if (npoints >= ptnumbuf) {
	    if ((ptbuf = (VERTEX *)
		 realloc(ptbuf, (ptnumbuf+N)*sizeof(VERTEX))) == NULL) {
		fprintf(info, "Error mallocing polygon memory\n");
		exit(1);
	    }
	    ptnumbuf += N;
	}
	pt = ptbuf + npoints;
	if (sscanf (line,"%lf %lf",&pt->lat,&pt->lon) != 2) goto poly_err;
	npoints++;
    }
    *p_pts = ptbuf;
    return (npoints);

 poly_err:
    fprintf(info, "Error reading box file\n");
    exit(1);
}

#define ordered(x1,x2,x3)    (((x1>=x2 && x2>=x3)) || ((x1<=x2) && (x2<=x3)))

/************************************************************************/
/* within_polygon -							*/
/*	Determine whether point is inside or outside of closed polygon.	*/
/*	Return 1 if inside, 0 if outside.				*/
/************************************************************************/
int within_polygon (VERTEX *p_pts, int npoints, VERTEX *pt)
{
    VERTEX pexterior;
    VERTEX *p1,*p2;
    VERTEX *pe = &pexterior;
    double a, b;
    double x;
    int i;
    int intersects = 0;

    /* Use a simple line-crossing algorithm to determine whether point	*/
    /* is interior.  Create a line segment between the point and any	*/
    /* point known to be exterior to the polygon.  Compute the number	*/
    /* of intersections between this line segment and the edges of the	*/
    /* polygon.  If the number of intersections is odd, the point is	*/
    /* within the polygon.  If the number of intersections is even, the	*/
    /* point is outside.						*/

    /* Assumptions on this implementation:				*/
    /*	1.  We work on a mercator projection with bounding box:		*/
    /*	    (90,-180),(90,180),(-90,180),(-90,-180).			*/
    /*	2.  Polygon does not wrap from top-bottom or left-right.	*/

    pexterior.lat = pt->lat;
    pexterior.lon = 200;
    for (i=0; i < npoints; i++) {
	/* Get endpoints of the edge of the polygon.			*/
	p1 = p_pts+i;
	p2 = (i < npoints - 1) ? p1+1 : p_pts;
	/* Skip degenerate edge (ie point).				*/
	if (p1->lat == p2->lat && p1->lon == p2->lon) continue;
	/* Skip edges clearly above, below, left, or right of test line.*/
	if ((p1->lat > pt->lat && p2->lat > pt->lat) ||
	    (p1->lat < pt->lat && p2->lat < pt->lat)) continue;
	if ((p1->lon < pt->lon && p2->lon < pt->lon) ||
	    (p1->lon > pe->lon && p2->lon > pe->lon)) continue;
	/* Test for horizontal edge segment containing point.		*/
	if (p1->lat == p2->lat) {
	    if ((p1->lat == pt->lat) && ordered(p1->lat,pt->lat,p2->lat))
		++intersects;
	    continue;
	}
	/* Test for vertical edge segment containing point.		*/
	if (p1->lon == p2->lon) {
	    if (ordered(pt->lon,p1->lon,pe->lon))
		++intersects;
	    continue;
	}
	/* We have a possible intersection.  Compute the intersection.	*/
	/* Compute a,b for line y=ax+b containing [p1,p2].		*/
	/* We know that the line is not vertical by previous tests.	*/
	a = (p1->lat - p2->lat) / (p1->lon - p2->lon);
	b = p1->lat - (a * p1->lon);
	/* Determine x value of intersect of horizontal [pt,pe] line.	*/
	x = (pt->lat - b) / a;
	/* If the x value is withing the [pt,pe] line segment, we have	*/
	/* an intersection.						*/
	if ordered(pt->lon,x,pe->lon) ++intersects;
    }
    return ((intersects % 2));
}

/************************************************************************/
/*  fgetline -								*/
/*	Read the next non-comment line in the file.  Return NULL on EOF.*/
/*	If flag == NEXT_LINE do not skip empty or comment lines.	*/
/************************************************************************/
static char *fgetline(char *line, int linelen, FILE *fp, int flag)
{
    char *p;
    line[linelen] = '\0';
    while (1) {
	if (fgets(line, linelen, fp)==NULL) return (NULL);
	if (flag == NEXT_LINE) return (line);
	/* Scan for first non-whitespace.				*/
	p = line;
	while (*p != 0 && isspace(*p)) p++;
	/* If first non-whitespace is not a comment, return line.	*/
	if (*p != 0 && *p != '#') return (line);
    }
}

