/************************************************************************/
/*  Routines for computing window times for events			*/
/*									*/
/*	Douglas Neuhauser						*/
/*	Seismological Laboratory					*/
/*	University of California, Berkeley				*/
/*	doug@seismo.berkeley.edu					*/
/*									*/
/************************************************************************/

/*
 * Copyright (c) 1996-2000 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

#ifndef lint
static char sccsid[] = "$Id: window.c,v 1.6 2003/10/29 21:17:28 lombard Exp $ ";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "qlib2.h"

#include "externals.h"
#include "procs.h"

#define LINELEN	    127

#define	Min_Speed_0	1.
#define	Min_Speed_1	2.5
#define	Max_Speed	8.5

static int comp_win_start
(const void       *w1, 		/* ptr window 1 for sorting.             */
 const void       *w2);		/* ptr window 2 for sorting.             */

/************************************************************************/
/* event_window:							*/
/*	Compute the P and S phase arrival times for an event		*/
/*	(lat,lon,time) at a given station.				*/
/************************************************************************/
static event_window 
(VERTEX	sta_loc, 	/* station location, lat and long	*/
 VERTEX	evt_loc,	/* event location, lat and long		*/
 double	*first,		/* first (P-wave) arrival time offset.	*/
 double	*last,		/* last (S-wave) arrival time offset.	*/
 double	*dist_km)	/* distance in km from event to station.*/
{
    double lat, lon, distance_km, tdif;

    /* Compute the first and last time for each channel based on	*/
    /* min and max speeds for wave propogation, epicenter, and station.	*/
    distance_km = distance (evt_loc, sta_loc);

    /* Compute start time. */	
    tdif = distance_km / Max_Speed;
    *first = tdif*USECS_PER_SEC;

    /* Compute end time. */
    if (distance_km <= 100) tdif = distance_km / Min_Speed_0;
    else tdif = distance_km / Min_Speed_1;
    *last = tdif*USECS_PER_SEC;
    if (debug(DEBUG_WINDOW)) {
	fprintf (info, "Event location = %.4lf %.4lf, Distance = %.2lf km, P = %.4lf, Sw = %.4lf\n",
		 evt_loc.lat, evt_loc.lon, 
		 distance_km, *first/USECS_PER_SEC, *last/USECS_PER_SEC);
    }
    *dist_km = distance_km;
    return;
}

/************************************************************************/
/* get_windows:								*/
/*	collect all the windows for events plus possibly the box window */
/*	and sort them by start time. Exits on errors.			*/
/************************************************************************/
void get_windows(EVENT *evts, int nevts, VERTEX *points, int npoints, 
		 WINDOW **p_windows, int *nwindows, WINDOW *box_window, 
		 DATA_HDR *hdr)
{
    VERTEX sta_loc = {999.0, 999.0};
    WINDOW *windows, *win;
    EVENT *evt;
    FILE *fp;
    INT_TIME t0, t1, t2, t3, tmin, tmax;
    EXT_TIME time_et;
    double lat, lon, w1, w2, dist_km;
    char date_str[16], start[16], end[16];
    char line[LINELEN+1], name[8];
    char *p, c;
    int nwinbuf = 0;
    int span_initialized;
    int i;
    
    /* Look up the station location. There is a time period in the 	*/
    /* coord file so we will use the first event time or the box start 	*/
    /* time to check against the coord-file time.			*/
    *nwindows = 0;
    if (nevts > 0) {
	time_et = int_to_ext(evts[0].time);
    }
    else if (box_window) {
	time_et = int_to_ext(box_window->start);
    }
    else {
	fprintf(info, "No events or box window\n");
	return;
    }

    if ((fp=fopen(coord_file,"r"))==NULL) {
	fprintf (info, "Error: Unable to open coord file %s\n", coord_file);
	exit(1);
    }
    sprintf(date_str, "%04d.%03d.%02d%02d", time_et.year, time_et.doy,
	    time_et.hour, time_et.minute);
    while(fgets(line,LINELEN,fp)) {
	if (sscanf(line, "%s %lf %lf %*f %s %s", name, &lat, &lon, start,
		   end) != 5) {
	    fprintf(info, "Error: Format error on line %d in coord file %s\n", i, coord_file);
	    fclose(fp);
	    exit(1);
	}
	if (strcasecmp(name,hdr->station_id)==0) {
	    if ( (strcmp(start, date_str) <= 0) &&
		 (strcmp(date_str, end) <= 0) )
		sta_loc.lat = lat;
	    sta_loc.lon = lon;
	    break;
	}
    }
    fclose(fp);
    if (sta_loc.lat > 90.) {
	fprintf (info, "Error: Missing or erroneous location data for %s\n", 
		 hdr->station_id);
	exit(1);
    }
    nwinbuf = nevts;
    if (box_window) nwinbuf++;
    if ((windows = (WINDOW *) malloc(nwinbuf * sizeof(WINDOW))) == NULL) {
	fprintf(info, "Error mallocing window buffer\n");
	exit(1);
    }
    *p_windows = windows;

    for (i = 0; i < nevts; i++) {
	evt = &evts[i];
	win = &windows[i];

	event_window( sta_loc, evt->loc, &w1, &w2, &dist_km);
	if (debug(DEBUG_WINDOW)) {
	    fprintf (info, "Event time     = %s\n", 
		     time_to_str(evt->time,date_fmt));
	}
	span_initialized = 0;
	for (p=window_selection; p!=NULL&&*p ;p++) {
	    c=*p;
	    switch (c) {
	    case 'p':
		/* Compute desired pre_p to post_s time window.	*/
		t1 = add_time(evt->time,0,(int)w1);
		t2 = add_time(evt->time,0,(int)w2);
		t0 = add_time(t1,-pre_p,0);
		t3 = add_time(t2,post_s,0);
		if (debug(DEBUG_WINDOW)) {
		    fprintf (info, "Phase window   = %s to ",
			     time_to_str(t0,date_fmt));
		    fprintf (info, "%s\n", time_to_str(t3,date_fmt));
		}
		break;
	    case 'f':
		/* Compute desired origin-based time window.	*/
		t0 = add_time(evt->time,-pre_origin,0);
		t3 = add_time(t0,origin_duration,0);
		if (debug(DEBUG_WINDOW)) {
		    fprintf (info, "Origin window  = %s to ",
			     time_to_str(t0,date_fmt));
		    fprintf (info, "%s\n", time_to_str(t3,date_fmt));
		}
		break;
	    default:
		fprintf (info, "Error: unknown window selection: %c\n", c);
		exit(1);
	    }
	    if (! span_initialized) {
		tmin = t0;
		tmax = t3;
		++span_initialized;
	    }
	    else {
		tmin = (tdiff(t0,tmin)<0) ? t0 : tmin;
		tmax = (tdiff(t3,tmax)>0) ? t3 : tmax;
	    }
	}
	win->start = tmin;
	win->end = tmax;
	win->event = evt;
	if (debug(DEBUG_WINDOW)) {
	    fprintf (info, "Extract window = %s to ",
		     time_to_str(tmin,date_fmt));
	    fprintf (info, "%s\n", time_to_str(tmax,date_fmt));
	    fflush (info);
	}
    }
    *nwindows = nevts;
    if (box_window) {
	if (within_polygon(points, npoints, &sta_loc)) {
	    windows[*nwindows] = *box_window;
	    (*nwindows)++;
	    if (debug(DEBUG_WINDOW))
		fprintf(info, "%s is within box\n", hdr->station_id);
	} else {
	    if (debug(DEBUG_WINDOW))
		fprintf(info, "%s is outside box\n", hdr->station_id);
	}
    }

    /* Sort the windows by start time */
    qsort(windows, *nwindows, sizeof(WINDOW), comp_win_start);

    return;
}


static int comp_win_start
(const void *w1,		/* ptr window 1 for sorting.             */
 const void *w2)		/* ptr window 2 for sorting.             */
{
    WINDOW       *win1, *win2;
    int result;
    double tdif;
    win1 = (WINDOW *)w1;
    win2 = (WINDOW *)w2;
    tdif = tdiff(win1->start, win2->start);
    if (tdif < 0.) result = -1;
    else if (tdif == 0) result = 0;
    else result = 1;
    return(result);
}
